*Clustering criterions
*SPSS macros by Kirill Orlov
*kior@comtv.ru; orlovk@ri-vita.ru
*http://ri-vita.ru/spssmacros.shtml
*All rights reserved.

*/*!CALHARV*/***************************************************************************************.
define !calharv(vars= !charend('/') /missing= !charend('/') /clusol= !charend('/'))
set results none.
!if (!missing='VARIABLE') !then !let !miss= '-999' !else !let !miss= 'omit' !ifend
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get vars /vari= !clu !vars /miss= !miss.
-do if cmin(vars(:,1))=cmax(vars(:,1)).
- comp calhar= -999.
-else.
- comp clu= design(vars(:,1)).
- comp vars= vars(:,2:ncol(vars)).
- comp valid= vars<>-999.
- comp vars= vars&*valid.
- comp n= t(valid)*clu.
- comp k= rsum(n>0).
- do if any(k<2).
-  comp calhar= -999.
- else.
-  comp nt= rsum(n).
-  comp sum= t(vars)*clu.
-  comp ss= t(vars&*vars)*clu.
-  comp sum2= sum&*sum.
-  comp ssw= rsum(ss-sum2/(n+(n=0))).
-  comp ssb= rsum(sum2/(n+(n=0)))-rsum(sum)&**2/nt.
-  comp calhar= csum(ssb&*(nt-k))/csum(ssw&*(k-1)). 
- end if.
-end if.
-save {!quote(!clu),calhar} /out= !quote(!conc('temp_calharv_',!clu,'.sav')) /vari= clusol calhar /strings= clusol.
-end matrix.
!doend
filter off.
set results list.
add files /file= !quote(!conc('temp_calharv_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_calharv_',!clu,'.sav')).
!doend
recode calhar (-999=sysmis).
cache.
exec.
var lab calhar 'Calinski-Harabasz' clusol 'Cluster solution'.
var lev calhar (sca).
format calhar (f8.4).
list clusol calhar.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(calhar) BY clusol /TITLE= 'Comparison of cluster solutions (Calinski-Harabasz)' /FOOTNOTE= 'The higher the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file !quote(!conc('temp_calharv_',!clu,'.sav')).
!doend
!enddefine.

*/*!CALHARM*/***************************************************************************************.
define !calharm(matrix= !charend('/') /square= !charend('/') /clusol= !charend('/'))
set mxloops 100000 results none.
comp casenum#= $casenum.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get d /vari= casenum# !matrix.
-comp d= d(:,d(:,1)+1).
 !if (!square='YES') !then
- comp d= d&*d.
 !ifend
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp calhar= -999.
-else.
- comp nt= nrow(clu).
- comp sst= msum(d)/2/nt.
- comp clu= design(clu).
- comp n= csum(clu).
- comp ssw= 0.
- loop i= 1 to ncol(clu).
-  do if n(i)>1.
-   comp c= kroneker(clu(:,i),make(1,nt,1)).
-   comp c= c&*t(c).
-   comp ssw= ssw+msum(d&*c)/2/n(i).
-  end if.
- end loop.
- comp calhar= (sst-ssw)/ssw*(nt-ncol(clu))/(ncol(clu)-1).
-end if.
-save {!quote(!clu),calhar} /out= !quote(!conc('temp_calharm_',!clu,'.sav')) /vari= clusol calhar /string= clusol.
-end matrix.
!doend
filter off.
del var casenum#.
set results list.
add files /file= !quote(!conc('temp_calharm_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_calharm_',!clu,'.sav')).
!doend
recode calhar (-999=sysmis).
cache.
exec.
var lab calhar 'Calinski-Harabasz' clusol 'Cluster solution'.
var lev calhar (sca).
format calhar (f8.4).
list clusol calhar.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(calhar) BY clusol /TITLE= 'Comparison of cluster solutions (Calinski-Harabasz)' /FOOTNOTE= 'The higher the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file !quote(!conc('temp_calharm_',!clu,'.sav')).
!doend
!enddefine.

*/*!DAVBOULV*/***************************************************************************************.
define !davboulv(vars= !charend('/') /clusol= !charend('/'))
set mxloops 100000 results none.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get vars /vari= !clu !vars /miss= omit.
-do if cmin(vars(:,1))=cmax(vars(:,1)).
- comp davboul= -999.
-else.
- comp clu= design(vars(:,1)).
- comp vars= vars(:,2:ncol(vars)).
- comp k= ncol(clu).
- comp n= kroneker(csum(clu),make(ncol(vars),1,1)).
- comp sum= t(vars)*clu.
- comp ss= t(vars&*vars)*clu.
- release vars.
- comp ss= ss-sum&*sum/n.
- comp ss= sqrt(csum(ss)/n(1,:)).
- comp sum= sum/n.
- comp dc= sscp(sum).
- comp @hsq= diag(dc)*make(1,ncol(dc),1).
- comp dc= sqrt(@hsq+t(@hsq)-2*dc).
- call setdiag(dc,1).
- do if not all(dc).
-  comp dc= dc+.000000001.
- end if.
- comp davboul= make(k,1,0).
- loop i= 1 to k.
-  loop j= 1 to k.
-   do if i<>j.
-    comp davboul(i)= rmax({davboul(i),(ss(i)+ss(j))/dc(i,j)}).
-   end if.
-  end loop.
- end loop.
- comp davboul= csum(davboul)/k.
-end if.
-save {!quote(!clu),davboul} /out= !quote(!conc('temp_davboulv_',!clu,'.sav')) /vari= clusol davboul /string= clusol.
-end matrix.
!doend
filter off.
set results list.
add files /file= !quote(!conc('temp_davboulv_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_davboulv_',!clu,'.sav')).
!doend
recode davboul (-999=sysmis).
cache.
exec.
var lab davboul 'Davies-Bouldin' clusol 'Cluster solution'.
var lev davboul (sca).
format davboul (f8.4).
list clusol davboul.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(davboul) BY clusol /TITLE= 'Comparison of cluster solutions (Davies-Bouldin)' /FOOTNOTE= 'The lower the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file !quote(!conc('temp_davboulv_',!clu,'.sav')).
!doend
!enddefine.

*/*!DAVBOULM*/***************************************************************************************.
define !davboulm(matrix= !charend('/') /square= !charend('/') /clusol= !charend('/'))
set mxloops 100000 results none.
comp casenum#= $casenum.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get d /vari= casenum# !matrix.
-comp d= d(:,d(:,1)+1).
 !if (!square='YES') !then
- comp d= d&*d.
 !ifend
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp davboul= -999.
-else.
- comp clu= design(clu).
- comp k= ncol(clu).
- comp n= csum(clu).
- comp ss = make(1,k,0).
- loop i= 1 to k.
-  comp sel= clu(:,i).
-  comp ss(i)= msum(d&*(sel*t(sel))).
- end loop.
- comp ss= ss/n.
- comp dc= make(k,k,0).
- loop i= 1 to k-1.
-  loop j= i+1 to k.
-   comp sel= clu(:,i)+clu(:,j).
-   comp dc(i,j)= msum(d&*(sel*t(sel)))-(n(i)+n(j))*(ss(i)+ss(j)).
-  end loop.
- end loop.
- comp dc= sqrt(dc/(2*t(n)*n)).
- comp dc= dc+t(dc).
- comp ss= sqrt(ss/(2*n)).
- call setdiag(dc,1).
- do if not all(dc).
-  comp dc= dc+.000000001.
- end if.
- comp davboul= make(k,1,0).
- loop i= 1 to k.
-  loop j= 1 to k.
-   do if i<>j.
-    comp davboul(i)= rmax({davboul(i),(ss(i)+ss(j))/dc(i,j)}).
-   end if.
-  end loop.
- end loop.
- comp davboul= csum(davboul)/k.
-end if.
-save {!quote(!clu),davboul} /out= !quote(!conc('temp_davboulm_',!clu,'.sav')) /vari= clusol davboul /string= clusol.
-end matrix.
!doend
filter off.
del var casenum#.
set results list.
add files /file= !quote(!conc('temp_davboulm_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_davboulm_',!clu,'.sav')).
!doend
recode davboul (-999=sysmis).
cache.
exec.
var lab davboul 'Davies-Bouldin' clusol 'Cluster solution'.
var lev davboul (sca).
format davboul (f8.4).
list clusol davboul.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(davboul) BY clusol /TITLE= 'Comparison of cluster solutions (Davies-Bouldin)' /FOOTNOTE= 'The lower the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file !quote(!conc('temp_davboulm_',!clu,'.sav')).
!doend
!enddefine.

*/*!RATLAN*/***************************************************************************************.
define !ratlan(vars= !charend('/') /missing= !charend('/') /clusol= !charend('/'))
set results none.
!if (!missing='VARIABLE') !then !let !miss= '-999' !else !let !miss= 'omit' !ifend
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get vars /vari= !clu !vars /miss= !miss.
-do if cmin(vars(:,1))=cmax(vars(:,1)).
- comp ratklan= -999.
-else.
- comp clu= design(vars(:,1)).
- comp vars= vars(:,2:ncol(vars)).
- comp valid= vars<>-999.
- comp vars= vars&*valid.
- comp n= t(valid)*clu.
- comp k= rsum(n>0).
- do if any(k<2).
-  comp ratklan= -999.
- else.
-  comp nt= rsum(n).
-  comp sum= t(vars)*clu.
-  comp ss= t(vars&*vars)*clu.
-  comp ssb= rsum(sum&*sum/(n+(n=0)))-rsum(sum)&**2/nt.
-  comp sst= rsum(ss)-rsum(sum)&**2/nt.
-  comp ratklan= csum(sqrt(ssb/sst))/ncol(vars). 
- end if.
-end if.
-save {!quote(!clu),ratklan} /out= !quote(!conc('temp_ratklan_',!clu,'.sav')) /vari= clusol ratklan /strings= clusol.
-end matrix.
!doend
filter off.
set results list.
add files /file= !quote(!conc('temp_ratklan_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_ratklan_',!clu,'.sav')).
!doend
recode ratklan (-999=sysmis).
cache.
exec.
var lab ratklan 'Ratkowsky-Lance' clusol 'Cluster solution'.
var lev ratklan (sca).
format ratklan (f8.4).
list clusol ratklan.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(ratklan) BY clusol /TITLE= 'Comparison of cluster solutions (Ratkowsky-Lance)' /FOOTNOTE= 'The higher the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file !quote(!conc('temp_ratklan_',!clu,'.sav')).
!doend
!enddefine.

*/*!RPBCLU*/***************************************************************************************.
define !rpbclu(matrix= !charend('/') /clusol= !charend('/'))
set results none.
comp casenum#= $casenum.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get d /vari= casenum# !matrix.
-comp d= d(:,d(:,1)+1).
-do if d(1,1).
- call setdiag(d,0).
- comp d= -d.
- comp d= d-mmin(d).
-end if.
-call setdiag(d,0).
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp r= -999.
-else.
- comp clu= kroneker(make(1,nrow(clu),1),clu). 
- comp clu= (clu=t(clu)).
- call setdiag(clu,0).
- comp sw= msum(d&*clu)/2.
- comp sb= msum(d)/2-sw.
- comp nw= msum(clu).
- comp nb= nrow(clu)**2-nrow(clu)-nw.
- comp ssdev= sqrt((mssq(d)-msum(d)&**2/(nw+nb))/2).
- comp r= (sb/nb-sw/nw)/ssdev*sqrt(nb*nw*2/(nb+nw)).
-end if.
-save {!quote(!clu),r} /out= !quote(!conc('temp_rpbclu_',!clu,'.sav')) /vari= clusol r_pb /strings= clusol.
-end matrix.
!doend
filter off.
del var casenum#.
set results list.
add files /file= !quote(!conc('temp_rpbclu_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_rpbclu_',!clu,'.sav')).
!doend
recode r_pb (-999=sysmis).
cache.
exec.
var lab r_pb 'Point-biserial r' clusol 'Cluster solution'.
var lev r_pb (sca).
format r_pb (f8.4).
list clusol r_pb.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(r_pb) BY clusol /TITLE= 'Comparison of cluster solutions (r)' /FOOTNOTE= 'The higher the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file= !quote(!conc('temp_rpbclu_',!clu,'.sav')).
!doend
!enddefine.

*/*!GAMMACLU*/***************************************************************************************.
define !gammaclu(matrix= !charend('/') /clusol= !charend('/'))
set mxloops 100000 results none.
comp casenum#= $casenum.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get d /vari= casenum# !matrix.
-comp d= d(:,d(:,1)+1).
-do if not d(1,1).
- comp d= -d.
-end if.
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp gamma= -999.
-else.
- comp clu= kroneker(make(1,nrow(clu),1),clu). 
- comp clu= (clu=t(clu)).
- comp x= make((ncol(d)**2-ncol(d))/2,1,0).
- comp y= x.
- comp ind= 0.
- loop i= 1 to ncol(d)-1.
-  loop j= i+1 to ncol(d).
-   comp ind= ind+1.
-   comp x(ind)= d(i,j).
-   comp y(ind)= clu(i,j).
-  end loop.
- end loop.
- release d,clu.
- comp rdes= design(x).
- comp cdes= design(y).
- comp f= t(rdes)*cdes.
- comp nr= ncol(rdes).
- comp nc= ncol(cdes).
- comp list= t(csum(kroneker(x,make(1,nr,1))&*rdes)/csum(rdes)).
- comp f(grade({list,f}(:,1)),:)= f.
- comp list= csum(kroneker(y,make(1,nc,1))&*cdes)/csum(cdes).
- comp f(:,grade({list;f}(1,:)))= f.
- release x,y,rdes,cdes,list.
- comp c= make(nr,nc,0).
- comp c(1:(nr-1),1:(nc-1))= f(2:nr,2:nc).
- loop j= nc-1 to 1 by -1.
-  comp c(:,j)= rsum(c(:,j:(j+1))).
- end loop.
- loop i= nr-1 to 1 by -1.
-  comp c(i,:)= csum(c(i:(i+1),:)).
- end loop.
- comp p= msum(f&*c).
- comp c= make(nr,nc,0).
- comp c(1:(nr-1),2:nc)= f(2:nr,1:(nc-1)).
- loop j= 2 to nc.
-  comp c(:,j)= rsum(c(:,(j-1):j)).
- end loop.
- loop i= nr-1 to 1 by -1.
-  comp c(i,:)= csum(c(i:(i+1),:)).
- end loop.
- comp q= msum(f&*c).
- comp gamma= (p-q)/(p+q).
-end if.
-save {!quote(!clu),gamma} /out= !quote(!conc('temp_gammaclu_',!clu,'.sav')) /vari= clusol gamma /strings= clusol.
-end matrix.
!doend
filter off.
del var casenum#.
set results list.
add files /file= !quote(!conc('temp_gammaclu_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_gammaclu_',!clu,'.sav')).
!doend
recode gamma (-999=sysmis).
cache.
exec.
var lab gamma 'Gamma' clusol 'Cluster solution'.
var lev gamma (sca).
format gamma (f8.4).
list clusol gamma.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(gamma) BY clusol /TITLE= 'Comparison of cluster solutions (Gamma)' /FOOTNOTE= 'The higher the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file= !quote(!conc('temp_gammaclu_',!clu,'.sav')).
!doend
!enddefine.

*/*!CINDEX*/***************************************************************************************.
define !cindex(matrix= !charend('/') /clusol= !charend('/'))
set mxloops 100000 results none.
comp casenum#= $casenum.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get d /vari= casenum# !matrix.
-comp d= d(:,d(:,1)+1).
-do if d(1,1).
- call setdiag(d,0).
- comp d= -d.
- comp d= d-mmin(d).
-end if.
-call setdiag(d,0).
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp cindex= -999.
-else.
- comp clu= kroneker(make(1,nrow(clu),1),clu). 
- comp clu= (clu=t(clu)).
- call setdiag(clu,0).
- comp sw= msum(d&*clu)/2.
- comp nw= msum(clu).
- comp nb= nrow(clu)**2-nrow(clu)-nw.
- comp dvec= make((nrow(d)**2-nrow(d))/2,1,0).
- comp elem= 0.
- loop i= 1 to nrow(d)-1.
-  loop j= i+1 to nrow(d).
-   comp elem= elem+1.
-   comp dvec(elem)= d(i,j).
-  end loop.
- end loop.
- release d.
- comp rdvec= grade(dvec).
- comp minsw= csum(dvec&*(rdvec<=nw/2)).
- comp maxsw= csum(dvec&*(rdvec>nb/2)).
- comp cindex= (sw-minsw)/(maxsw-minsw).
-end if.
-save {!quote(!clu),cindex} /out= !quote(!conc('temp_cindex_',!clu,'.sav')) /vari= clusol c_index /strings= clusol.
-end matrix.
!doend
filter off.
del var casenum#.
set results list.
add files /file= !quote(!conc('temp_cindex_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_cindex_',!clu,'.sav')).
!doend
recode c_index (-999=sysmis).
cache.
exec.
var lab c_index 'C-Index' clusol 'Cluster solution'.
var lev c_index (sca).
format c_index (f8.4).
list clusol c_index.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(c_index) BY clusol /TITLE= 'Comparison of cluster solutions (C-Index)' /FOOTNOTE= 'The lower the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file= !quote(!conc('temp_cindex_',!clu,'.sav')).
!doend
!enddefine.

*/*!DUNN*/***************************************************************************************.
define !dunn(matrix= !charend('/') /clusol= !charend('/'))
set results none.
comp casenum#= $casenum.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get d /vari= casenum# !matrix.
-comp d= d(:,d(:,1)+1).
-do if d(1,1).
- call setdiag(d,0).
- comp d= -d.
- comp d= d-mmin(d).
-end if.
-call setdiag(d,0).
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp dunn= -999.
-else.
- comp clu= kroneker(make(1,nrow(clu),1),clu). 
- comp clu= (clu=t(clu)).
- comp maxw= mmax(d&*clu).
- comp minb= mmin(d&*(not clu)+clu*(mmax(d)+1)).
- do if maxw=0.
-  comp dunn= -999.
- else.
-  comp dunn= minb/maxw.
- end if.
-end if.
-save {!quote(!clu),dunn} /out= !quote(!conc('temp_dunn_',!clu,'.sav')) /vari= clusol dunn /strings= clusol.
-end matrix.
!doend
filter off.
del var casenum#.
set results list.
add files /file= !quote(!conc('temp_dunn_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_dunn_',!clu,'.sav')).
!doend
recode dunn (-999=sysmis).
cache.
exec.
var lab dunn 'Dunn' clusol 'Cluster solution'.
var lev dunn (sca).
format dunn (f8.4).
list clusol dunn.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(dunn) BY clusol /TITLE= 'Comparison of cluster solutions (Dunn)' /FOOTNOTE= 'The higher the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file= !quote(!conc('temp_dunn_',!clu,'.sav')).
!doend
!enddefine.

*/*!SILHOU*/***************************************************************************************.
define !silhou(matrix= !charend('/') /clusol= !charend('/') /type= !charend('/') /caps= !charend('/') /single= !charend('/'))
!let !bclu= !unquote(!head(!caps)) !let !sil= !unquote(!head(!tail(!caps)))
!if (!single='SYSMIS') !then !let !sing= !single !else !let !sing= 0 !ifend
set mxloops 100000 results none errors none.
comp casenum#= $casenum.
!do !clu !in (!clusol)
-filter by !clu.
-matrix.
-get d /vari= casenum# !matrix.
-comp case= d(:,1).
-comp d= d(:,d(:,1)+1).
-do if d(1,1).
- call setdiag(d,0).
- comp d= -d.
- comp d= d-mmin(d).
-end if.
-call setdiag(d,0).
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp bclu= make(nrow(clu),1,-999).
- comp sil= make(nrow(clu),1,-999).
-else.
- comp des= design(clu).
- comp n= csum(des).
- comp abmx= make(nrow(des),ncol(des),0).
  !if (!type='AVER' !or !type=!null) !then
-  loop i= 1 to ncol(des).
-   comp abmx(:,i)= rsum(d&*t(kroneker(des(:,i),make(1,nrow(des),1)))).
-  end loop.
-  comp a= rsum(abmx&*des/kroneker(n-(n>1),make(nrow(des),1,1))).
-  comp abmx= abmx&*(not des)/kroneker(n,make(nrow(des),1,1)).
-  comp abmx= abmx+des*(mmax(abmx)+1).
  !ifend
  !if (!type='FARTH') !then
-  loop i= 1 to ncol(des).
-   comp abmx(:,i)= rmax(d&*t(kroneker(des(:,i),make(1,nrow(des),1)))).
-  end loop.
-  comp a= rsum(abmx&*des).
-  comp abmx= abmx&*(not des)+des*(mmax(abmx)+1).
  !ifend
  !if (!type='NEAR') !then
-  comp bigval= mmax(d)+1.
-  call setdiag(d,bigval).
-  loop i= 1 to ncol(des).
-   comp krdesi= t(kroneker(des(:,i),make(1,nrow(des),1))).
-   comp abmx(:,i)= rmin(d&*krdesi+(krdesi=0)*bigval).
-  end loop.
-  comp abmx= abmx&*(abmx<>bigval).
-  comp a= rsum(abmx&*des).
-  comp abmx= abmx&*(not des)+des*(mmax(abmx)+1).
  !ifend
  !if (!type='DEVIAT') !then
-  comp d= d&*d.
-  comp ones= make(1,ncol(d),1).
-  comp abmx= t(abmx).
-  loop i= 1 to ncol(des).
-   comp sel= des(:,i).
-   comp bsum= 2*csum(d&*(sel*ones)).
-   comp wsum= msum(d&*(sel*t(sel))).
-   comp abmx(i,:)= (bsum/n(i)-make(1,ncol(d),wsum/n(i)**2))/2.
-  end loop.
-  comp abmx= t(sqrt(abmx)).
-  comp a= rsum(abmx&*des).
-  comp abmx= abmx&*(not des)+des*(mmax(abmx)+1).
  !ifend
- comp b= rmin(abmx).
- comp bclu= make(nrow(clu),1,0).
- loop i= 1 to ncol(des).
-  comp bclu= bclu+((abmx(:,i)=b)*csum(des(:,i)&*clu)/n(i))&*(bclu=0).
- end loop.
- comp sil= (b-a)/(rmax({a,b})+(rmax({a,b})=0))+rsum(des&*kroneker(n=1,make(nrow(des),1,1))).
-end if.
-save {case,bclu,sil} /out= !quote(!conc('temp_silhou_',!clu,'.sav')) /vari= casenum# bclu sil.
-end matrix.
-del vari !conc(!bclu,!clu) !conc(!sil,!clu).
!doend
filter off.
set results list errors list.
!if (!type=!null) !then !let !tp= 'type=AVER' !else !let !tp= !conc('type=',!type) !ifend
!do !clu !in (!clusol)
-match files file= * /file= !quote(!conc('temp_silhou_',!clu,'.sav')) /rename= (bclu sil = !conc(!bclu,!clu) !conc(!sil,!clu)) /by casenum#.
-recode !conc(!sil,!clu) (2=!sing) (-999=sysmis) /!conc(!bclu,!clu) (-999=sysmis).
-var lab !conc(!bclu,!clu) !quote(!conc('Closest cluster, ',!tp)).
-var lev !conc(!bclu,!clu) (nom) /!conc(!sil,!clu) (sca).
!doend
cache.
exec.
del var casenum#.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= MEAN(!do !clu !in (!clusol) !conc(!sil,!clu) !doend) /INTERVAL SD(1) /MISSING= VARIABLE
       /TITLE= !quote(!conc('Comparison of cluster solutions by Silhouette Statistic ',!tp)) /FOOTNOTE= 'The higher the mean the "better" is solution'.
!ifend
!if (!type='DEVIAT') !then
-echo '>REMINDER. With TYPE=DEVIAT input matrix must be of non-squared euclidean'.
-echo ' distances or of distances that approach geometrically to being such.'.
-echo " Only in this case the macro's results make sense, for then one could speak".
-echo ' of cluster centroids.'.
!ifend
!do !clu !in (!clusol)
-erase file= !quote(!conc('temp_silhou_',!clu,'.sav')).
!doend
!enddefine.

*/*!AICBIC*/***************************************************************************************.
define !aicbic(scavars= !charend('/') /catvars= !charend('/') /clusol= !charend('/'))
!let !vlist= !null
!do !var !in (!conc(!scavars,' ',!catvars))
 !if (!upcase(!var)= 'TO') !then
  !let !vlist= !substr(!conc(' ',!vlist),1,!len(!vlist)) !let !vlist= !conc(!vlist,' to ')
 !else
  !let !vlist= !conc(!vlist,!var,',')
 !ifend
!doend
!let !vlist= !substr(!conc(' ',!vlist),1,!len(!vlist))
set mxloops 100000 results none.
!do !clu !in (!clusol)
-filter by !clu.
-temp.
-select if nmiss(!vlist)=0.
-matrix.
-get clu /var= !clu.
-do if cmin(clu)=cmax(clu).
- comp aic= -999.
- comp bic= -999.
-else.
- comp clu= design(clu).
- comp n= csum(clu).
- comp sloglik= 0.
- comp cloglik= 0.
- comp sm= 0.
- comp cm= 0.
  !if (!scavars<>!null) !then
-  get vars /vari= !scavars. 
-  comp nmx= kroneker(n,make(ncol(vars),1,1)).
-  comp sum= t(vars)*clu.
-  comp ss= t(vars&*vars)*clu.
-  comp vrnc= (ss-sum&*sum/nmx)/nmx.
-  comp vrnct= (t(cssq(vars))-t(csum(vars))&**2/nrow(clu))/(nrow(clu)-1).
-  comp vrnct= kroneker(vrnct,make(1,ncol(clu),1)).
-  comp sloglik= csum(ln(vrnc+vrnct)/2).
-  comp sm= 2*ncol(vars).
  !ifend
  !if (!catvars<>!null) !then
-  get catvars /vari= !catvars.
-  comp ncat= make(1,ncol(catvars),0).
-  loop i= 1 to ncol(catvars).
-   comp cat= design(catvars(:,i)).
-   comp ncat(i)= ncol(cat).
-   comp cat= t(cat)*clu/kroneker(n,make(ncol(cat),1,1)).
-   comp cloglik= cloglik-csum(cat&*ln(cat+(cat=0))).
-  end loop.
-  comp cm= rsum(ncat-1).
  !ifend
- comp loglik= -n&*(sloglik+cloglik).
- comp m= ncol(clu)*(sm+cm).
- comp aic= -2*rsum(loglik)+2*m.
- comp bic= -2*rsum(loglik)+m*ln(nrow(clu)).
-end if.
-save {!quote(!clu),aic,bic} /out= !quote(!conc('temp_aicbic_',!clu,'.sav')) /vari= clusol aic bic /strings= clusol.
-end matrix.
!doend
filter off.
set results list.
add files /file= !quote(!conc('temp_aicbic_',!head(!clusol),'.sav')). 
!do !clu !in (!tail(!clusol))
-add files /file= * /file= !quote(!conc('temp_aicbic_',!clu,'.sav')).
!doend
recode aic bic (-999=sysmis).
cache.
exec.
var lab aic 'Akaike Information Criterion' bic 'Bayesian Information Criterion' clusol 'Cluster solution'.
var lev aic bic (sca).
format aic bic (f8.4).
list clusol aic bic.
!if (!tail(!clusol)<>!null) !then
-GRAPH /LINE(SIMPLE)= VALUE(aic) BY clusol /TITLE= 'Comparison of cluster solutions (AIC)' /FOOTNOTE= 'The lower the value the "better" is solution'.
-GRAPH /LINE(SIMPLE)= VALUE(bic) BY clusol /TITLE= 'Comparison of cluster solutions (BIC)' /FOOTNOTE= 'The lower the value the "better" is solution'.
!ifend
!do !clu !in (!clusol)
-erase file !quote(!conc('temp_aicbic_',!clu,'.sav')).
!doend
!enddefine.
